import { Observable } from 'rxjs';
import { map } from 'rxjs/operators';
import { CallHandler, ExecutionContext, Injectable, NestInterceptor } from '@nestjs/common';

import { EntityType } from '../types/entity.type';
import { Serializable } from '../interfaces/serializable.interface';
import { ResourceSerializer } from '../serializers/resource.serializer';
import { Document, JSONAPIObject, MetaObject, ResourceIdentifierObject, ResourceObject } from '../interfaces/document.interface';

export interface EntitySerializerOptions<T extends Serializable> {
  entity: EntityType<T>;
  jsonapi?: {
    meta?: MetaObject;
    include?: boolean;
  };
}

@Injectable()
export class EntitySerializerInterceptor<T extends Serializable> implements NestInterceptor<T, Document<T>> {
  constructor(private options: EntitySerializerOptions<T>) { }

  public intercept(context: ExecutionContext, next: CallHandler): Observable<Document<T>> {
    return next.handle().pipe(map(data => { 
      const document: Document<T> = {};

      document.data = this.createData(data);

      if (this.options.jsonapi?.include !== false) {
        document.jsonapi = this.createJSONAPI();
      }
   
      return document;
    }));
  };

  private createData(data: any): ResourceObject<T> | ResourceObject<T>[] | ResourceIdentifierObject<T> | ResourceIdentifierObject<T>[] | null | [] {
    const serializer: ResourceSerializer<T> = new ResourceSerializer<T>(this.options);
    if (!Array.isArray(data)) {
      return serializer.parseOne(data);
    } else {
      return serializer.parseMany(data);
    }
  }

  private createJSONAPI(): JSONAPIObject {
    return {
      version: '1.0',
      meta: this.options.jsonapi?.meta,
    }
  }
}