import { Serializable } from './serializable.interface';

/**
 * JSONAPI - Document (Root Object)
 */
export type Document<T extends Serializable> = {
  data?: ResourceObject<T> | ResourceObject<T>[] | ResourceIdentifierObject<T> | ResourceIdentifierObject<T>[] | null | [];
  errors?: ErrorObject[];
  meta?: MetaObject;

  jsonapi?: JSONAPIObject;
  links?: Omit<LinksObject, 'about'>;
  included?: ResourceObject<any>[];
}

/**
 * JSONAPI - Meta Object
 */
export type MetaObject = {
  [key: string]: any;
}

/**
 * JSONAPI - Links Object
 */
export type LinksObject = {
  self?: string | LinkObject;
  related?: string | LinkObject;
  about?: string | LinkObject;
  first?: string | LinkObject;
  last?: string | LinkObject;
  prev?: string | LinkObject;
  next?: string | LinkObject;
}

export type LinkObject = {
  href?: string;
  meta?: MetaObject
}

/**
 * JSONAPI - Error Object
 */
export type ErrorObject = {
  id?: string;
  links?: Pick<LinksObject, 'about'>;
  status?: string;
  code?: string;
  title?: string;
  detail?: string;
  source?: SourceObject;
  meta?: MetaObject;
}

export type SourceObject = {
  pointer?: string;
  parameter?: string;
}

/**
 * JSONAPI - JSON:API Object
 */
export type JSONAPIObject = {
  version?: string;
  meta?: MetaObject
}

/**
 * JSONAPI - Resource Object
 */
export type ResourceObject<T extends Serializable> = {
  id: string | number;
  type: string;

  attributes?: AttributesObject<T>;
  relationships?: RelationshipsObject<T>;
  links?: Pick<LinksObject, 'self'>;
  meta?: MetaObject;
}

export type AttributesObject<T extends Serializable> = Partial<T>;

export type RelationshipsObject<T extends Serializable> = {
  [key: string]: RelationshipObject<T>;
}

export type RelationshipObject<T extends Serializable> = {
  links?: Pick<LinksObject, 'self' | 'related'>
  data?: ResourceLinkage<T>;
  meta?: MetaObject;
}

export type ResourceLinkage<T extends Serializable> = 
  ResourceIdentifierObject<T> | ResourceIdentifierObject<T>[] | null | [];

/**
 * JSONAPI - Resource Identifier Object
 */
export type ResourceIdentifierObject<T extends Serializable> = {
  id: string | number;
  type: string;
}