import { Serializable } from '../interfaces/serializable.interface';
import { EntitySerializerOptions } from '../interceptors/entity-serializer.interceptor';
import { ResourceIdentifierObject, ResourceObject } from '../interfaces/document.interface';
import { EntityOptions } from '../decorators/entity.decorator';
import { RelationshipOptions } from '../decorators/relationship.decorator';

export class ResourceSerializer<T extends Serializable> {
  constructor(private options: EntitySerializerOptions<T>) { }

  public parseOne(resource: T): ResourceObject<T> | ResourceIdentifierObject<T> | null {
    const id = resource.id;
    const type = this.contextEntity().name;

    // Attributes
    const attributes: Partial<T> = resource;
    delete attributes.id;
    
    // One to One Relationships
    Object.keys(this.contextOneToOne()).forEach((key: string) => {
      delete (attributes as any)[key];
    });

    // One to Many Relationships
    Object.keys(this.contextOneToMany()).forEach((key: string) => {
      delete (attributes as any)[key];
    });

    return { id, type, attributes };
  }

  public parseMany(resources: T[]): ResourceObject<T>[] | ResourceIdentifierObject<T>[] | [] {
    return resources.map((resource: T) => {
      const id = resource.id;
      const type = this.contextEntity().name;

      // Attributes
      const attributes: Partial<T> = resource;
      delete attributes.id;
      
      // One to One Relationships
      Object.keys(this.contextOneToOne()).forEach((key: string) => {
        delete (attributes as any)[key];
      });

      // One to Many Relationships
      Object.keys(this.contextOneToMany()).forEach((key: string) => {
        delete (attributes as any)[key];
      });

      return { id, type, attributes };
    });
  }

  private contextEntity(): EntityOptions {
    const entity: EntityOptions = Reflect.getMetadata('JSONAPI_ENTITY', this.options.entity);
    if (!entity) {
      throw new Error('JSON-API-Transformer could not resolve the entity type. Did you add an @Entity annotation?');
    } else {
      return entity;
    }
  }

  private contextOneToOne(): { [K in keyof T]: RelationshipOptions } {
    return Reflect.getMetadata('JSONAPI_ONE_TO_ONE', this.options.entity);
  }

  private contextOneToMany(): { [K in keyof T]: RelationshipOptions } {
    return Reflect.getMetadata('JSONAPI_ONE_TO_MANY', this.options.entity);
  }
}